@@ -6,25 +6,42 @@ import re
 import getopt
 import json
 
+import CloudFlare
+
+from .dump import dump_commands, dump_commands_from_web
+from . import converters
+from . import examples
+
 my_yaml = None
 my_jsonlines = None
 
-import CloudFlare
-from .dump import dump_commands, dump_commands_from_web
-from . import converters
+class CLI4InternalError(Exception):
+    """ errors in cli4 """
 
 def load_and_check_yaml():
     """ load_and_check_yaml() """
+    # only called if user uses --yaml flag
     from . import myyaml
     global my_yaml
-    my_yaml = myyaml.myyaml()
-    if not my_yaml.available():
-        sys.exit('cli4: install yaml support')
+    try:
+        my_yaml = myyaml.myyaml()
+    except ImportError:
+        sys.exit('cli4: install yaml support via: pip install pyyaml')
+
+def load_and_check_jsonlines():
+    """ load_and_check_yaml() """
+    # only called if user uses --ndjson flag
+    from . import myjsonlines
+    global my_jsonlines
+    try:
+        my_jsonlines = myjsonlines.myjsonlines()
+    except ImportError:
+        sys.exit('cli4: install jsonlines support via: pip install jsonlines')
 
 def strip_multiline(s):
     """ remove leading/trailing tabs/spaces on each line"""
     # This hack is needed in order to use yaml.safe_load() on JSON text - tabs are not allowed
-    return '\n'.join([l.strip() for l in s.splitlines()])
+    return '\n'.join([line.strip() for line in s.splitlines()])
 
 def process_params_content_files(method, binary_file, args):
     """ process_params_content_files() """
@@ -39,7 +56,7 @@ def process_params_content_files(method, binary_file, args):
     while len(args) > 0 and ('=' in args[0] or args[0][0] == '@'):
         arg = args.pop(0)
         if arg[0] == '@':
-            # a file to be uploaded - used in workers/script - only via PUT
+            # a file to be uploaded - used in workers/script etc - only via PUT or POST
             filename = arg[1:]
             if method not in ['PUT','POST']:
                 sys.exit('cli4: %s - raw file upload only with PUT or POST' % (filename))
@@ -54,7 +71,7 @@ def process_params_content_files(method, binary_file, args):
                         with open(filename, 'rb') as f:
                             content = f.read()
                     else:
-                        with open(filename, 'r') as f:
+                        with open(filename, 'r', encoding="utf-8") as f:
                             content = f.read()
             except IOError:
                 sys.exit('cli4: %s - file open failure' % (filename))
@@ -77,23 +94,26 @@ def process_params_content_files(method, binary_file, args):
         elif value_string[0] in '[{' and value_string[-1] in '}]':
             # a json structure - used in pagerules
             try:
-                #value = json.loads(value) - changed to yaml code to remove unicode string issues
+                # value = json.loads(value) - changed to yaml code to remove unicode string issues
                 load_and_check_yaml()
                 # cleanup string before parsing so that yaml.safe.load does not complain about whitespace
                 # >>> found character '\t' that cannot start any token <<<
                 value_string = strip_multiline(value_string)
                 try:
                     value = my_yaml.safe_load(value_string)
-                except my_yaml.parser.ParserError as e:
-                    raise ValueError
+                except my_yaml.parser.ParserError:
+                    raise ValueError from None
             except ValueError:
                 sys.exit('cli4: %s="%s" - can\'t parse json value' % (tag_string, value_string))
         elif value_string[0] == '@':
-            # a file to be uploaded - used in dns_records/import - only via POST
+            # a file to be uploaded - used in dns_records/import etc - only via PUT or POST
             filename = value_string[1:]
-            if method != 'POST':
-                sys.exit('cli4: %s=%s - file upload only with POST' % (tag_string, filename))
-            files = {}
+            if method not in ['PUT', 'POST']:
+                sys.exit('cli4: %s=%s - file upload only with PUT or POST' % (tag_string, filename))
+            if files is None:
+                files = {}
+            if tag_string in files:
+                sys.exit('cli4: %s=%s - duplicate name' % (tag_string, filename))
             try:
                 if filename == '-':
                     files[tag_string] = sys.stdin
@@ -103,6 +123,9 @@ def process_params_content_files(method, binary_file, args):
                 sys.exit('cli4: %s=%s - file open failure' % (tag_string, filename))
             # no need for param code below
             continue
+        elif (value_string[0] == '"' and value_string[-1] == '"') or (value_string[0] == '\'' and value_string[-1] == '\''):
+            # remove quotes
+            value = value_string[1:-1]
         else:
             value = value_string
 
@@ -123,6 +146,20 @@ def process_params_content_files(method, binary_file, args):
                 sys.exit('cli4: %s=%s - param error. Can\'t mix unnamed and named list' %
                          (tag_string, value_string))
 
+    if content and params:
+        sys.exit('cli4: content and params not allowed together')
+
+    if params and files:
+        for k,v in params.items():
+            files[k] = (None, v)
+        params = None
+        # sys.exit('cli4: params and files not allowed together')
+
+    if method != 'GET':
+        if params:
+            content = params
+            params = None
+
     return (params, content, files)
 
 def run_command(cf, method, command, params=None, content=None, files=None):
@@ -155,7 +192,7 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                 if len(element) in [32, 40, 48] and hex_only.match(element):
                     # raw identifier - lets just use it as-is
                     identifier1 = element
-                elif len(element) == 36  and uuid_value.match(element):
+                elif len(element) == 36 and uuid_value.match(element):
                     # uuid identifier - lets just use it as-is
                     identifier1 = element
                 elif element[0] == ':':
@@ -181,8 +218,8 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                         elif (cmd[0] == 'user') and (cmd[1] == 'load_balancers') and (cmd[2] == 'pools'):
                             identifier1 = converters.convert_load_balancers_pool_to_identifier(cf, element)
                         else:
-                            raise Exception("/%s/%s :NOT CODED YET" % ('/'.join(cmd), element))
-                    except Exception as e:
+                            raise CLI4InternalError("/%s/%s :NOT CODED YET" % ('/'.join(cmd), element))
+                    except CLI4InternalError as e:
                         sys.stderr.write('cli4: /%s - %s\n' % (command, e))
                         raise e
                 cmd.append(':' + identifier1)
@@ -190,7 +227,7 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                 if len(element) in [32, 40, 48] and hex_only.match(element):
                     # raw identifier - lets just use it as-is
                     identifier2 = element
-                elif len(element) == 36  and uuid_value.match(element):
+                elif len(element) == 36 and uuid_value.match(element):
                     # uuid identifier - lets just use it as-is
                     identifier2 = element
                 elif element[0] == ':':
@@ -199,16 +236,12 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                 else:
                     try:
                         if (cmd[0] and cmd[0] == 'zones') and (cmd[2] and cmd[2] == 'dns_records'):
-                            identifier2 = converters.convert_dns_record_to_identifier(cf,
-                                                                                      identifier1,
-                                                                                      element)
+                            identifier2 = converters.convert_dns_record_to_identifier(cf, identifier1, element)
                         elif (cmd[0] and cmd[0] == 'zones') and (cmd[2] and cmd[2] == 'custom_hostnames'):
-                            identifier2 = converters.convert_custom_hostnames_to_identifier(cf,
-                                                                                      identifier1,
-                                                                                      element)
+                            identifier2 = converters.convert_custom_hostnames_to_identifier(cf, identifier1, element)
                         else:
-                            raise Exception("/%s/%s :NOT CODED YET" % ('/'.join(cmd), element))
-                    except Exception as e:
+                            raise CLI4InternalError("/%s/:%s :NOT CODED YET" % ('/'.join(cmd), element))
+                    except CLI4InternalError as e:
                         sys.stderr.write('cli4: /%s - %s\n' % (command, e))
                         raise e
                 # identifier2 may be an array - this needs to be dealt with later
@@ -221,7 +254,7 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                 if len(element) in [32, 40, 48] and hex_only.match(element):
                     # raw identifier - lets just use it as-is
                     identifier3 = element
-                elif len(element) == 36  and uuid_value.match(element):
+                elif len(element) == 36 and uuid_value.match(element):
                     # uuid identifier - lets just use it as-is
                     identifier3 = element
                 elif waf_rules.match(element):
@@ -230,36 +263,28 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                     # raw string - used for workers script_names
                     identifier3 = element[1:]
                 else:
+                    #  /accounts/:id/storage/kv/namespaces/:id/values/:key_name - it's a strange one!
                     if len(cmd) >= 6 and cmd[0] == 'accounts' and cmd[2] == 'storage' and cmd[3] == 'kv' and cmd[4] == 'namespaces' and cmd[6] == 'values':
                         identifier3 = element
                     else:
-                        sys.stderr.write('/%s/%s :NOT CODED YET 3\n' % ('/'.join(cmd), element))
+                        sys.stderr.write('/%s/:%s :NOT CODED YET\n' % ('/'.join(cmd), element))
                         raise e
         else:
             try:
-                # dashes (vs underscores) cause issues in Python and other languages
-                if '-' in element:
-                    m = getattr(m, element.replace('-','_'))
-                else:
-                    m = getattr(m, element)
+                m = getattr(m, CloudFlare.CloudFlare.sanitize_verb(element))
                 cmd.append(element)
-            except AttributeError:
+            except AttributeError as e:
                 # the verb/element was not found
                 sys.stderr.write('cli4: /%s - not found\n' % (command))
                 raise e
 
-    if content and params:
-        sys.stderr.write('cli4: /%s - content and params not allowed together\n' % (command))
-        raise Exception
-    if content:
-        params = content
-
     results = []
     if identifier2 is None:
         identifier2 = [None]
     for i2 in identifier2:
         try:
             if method == 'GET':
+                # no content with a GET call
                 r = m.get(identifier1=identifier1,
                           identifier2=i2,
                           identifier3=identifier3,
@@ -268,22 +293,22 @@ def run_command(cf, method, command, params=None, content=None, files=None):
                 r = m.patch(identifier1=identifier1,
                             identifier2=i2,
                             identifier3=identifier3,
-                            data=params)
+                            data=content)
             elif method == 'POST':
                 r = m.post(identifier1=identifier1,
                            identifier2=i2,
                            identifier3=identifier3,
-                           data=params, files=files)
+                           data=content, files=files)
             elif method == 'PUT':
                 r = m.put(identifier1=identifier1,
                           identifier2=i2,
                           identifier3=identifier3,
-                          data=params)
+                          data=content, files=files)
             elif method == 'DELETE':
                 r = m.delete(identifier1=identifier1,
                              identifier2=i2,
                              identifier3=identifier3,
-                             data=params)
+                             data=content)
             else:
                 pass
         except CloudFlare.exceptions.CloudFlareAPIError as e:
@@ -312,9 +337,11 @@ def write_results(results, output):
     if len(results) == 1:
         results = results[0]
 
-    if isinstance(results, str):
+    if isinstance(results, (str, bytes, bytearray)):
         # if the results are a simple string, then it should be dumped directly
-        # this is only used for /zones/:id/dns_records/export presently
+        # this is only used for /zones/:id/dns_records/export, workers, and other calls
+        # or
+        # output is image or audio or video or something like that so we dump directly
         pass
     else:
         # anything more complex (dict, list, etc) should be dumped as JSON/YAML
@@ -342,14 +369,18 @@ def write_results(results, output):
                 pass
             return
         else:
-            # Internal error
-            pass
+            # None of the above, so pass thru results except something in byte form
+            if not isinstance(results, (bytes, bytearray)):
+                results = str(results)
 
     if results:
         try:
-            sys.stdout.write(results)
-            if not results.endswith('\n'):
-                sys.stdout.write('\n')
+            if isinstance(results, (bytes, bytearray)):
+                sys.stdout.buffer.write(results)
+            else:
+                sys.stdout.write(results)
+                if not results.endswith('\n'):
+                    sys.stdout.write('\n')
         except (BrokenPipeError, IOError):
             pass
 
@@ -358,37 +389,48 @@ def do_it(args):
 
     verbose = False
     output = 'json'
+    example = False
     raw = False
-    dump = False
-    dump_from_web = False
+    do_dump = False
+    do_openapi = None
     openapi_url = None
     binary_file = False
     profile = None
+    http_headers = None
+    warnings = True
     method = 'GET'
 
     usage = ('usage: cli4 '
-             + '[-V|--version] [-h|--help] [-v|--verbose] [-q|--quiet] '
-             + '[-j|--json] [-y|--yaml] [-n|ndjson] '
+             + '[-V|--version] [-h|--help] [-v|--verbose] '
+             + '[-e|--examples] '
+             + '[-q|--quiet] '
+             + '[-j|--json] [-y|--yaml] [-n|--ndjson] [-i|--image] '
              + '[-r|--raw] '
              + '[-d|--dump] '
-             + '[-a|--api] '
              + '[-A|--openapi url] '
              + '[-b|--binary] '
              + '[-p|--profile profile-name] '
+             + '[-h|--header additional-header] '
+             + '[-w|--warnings [True|False]] '
              + '[--get|--patch|--post|--put|--delete] '
              + '[item=value|item=@filename|@filename ...] '
              + '/command ...')
 
     try:
         opts, args = getopt.getopt(args,
-                                   'VhvqjyrdaA:bp:GPOUD',
+                                   'VhveqjynirdA:bp:h:w:GPOUD',
                                    [
-                                       'version',
-                                       'help', 'verbose', 'quiet', 'json', 'yaml', 'ndjson',
+                                       'version', 'help', 'verbose',
+                                       'examples',
+                                       'quiet',
+                                       'json', 'yaml', 'ndjson', 'image',
                                        'raw',
-                                       'dump', 'api', 'openapi=',
+                                       'dump',
+                                       'openapi=',
                                        'binary',
                                        'profile=',
+                                       'header=',
+                                       'warnings=',
                                        'get', 'patch', 'post', 'put', 'delete'
                                    ])
     except getopt.GetoptError:
@@ -402,28 +444,40 @@ def do_it(args):
             verbose = True
         elif opt in ('-q', '--quiet'):
             output = None
+        elif opt in ('-e', '--examples'):
+            example = True
         elif opt in ('-j', '--json'):
             output = 'json'
         elif opt in ('-y', '--yaml'):
             load_and_check_yaml()
             output = 'yaml'
         elif opt in ('-n', '--ndjson'):
-            from . import myjsonlines
-            global my_jsonlines
-            my_jsonlines = myjsonlines.myjsonlines()
-            if not my_jsonlines.available():
-                sys.exit('cli4: install jsonlines support')
+            load_and_check_jsonlines()
             output = 'ndjson'
+        elif opt in ('-i', '--image'):
+            output = 'image'
         elif opt in ('-r', '--raw'):
             raw = True
         elif opt in ('-p', '--profile'):
             profile = arg
+        elif opt in ('-h', '--header'):
+            if http_headers is None:
+                http_headers = []
+            http_headers.append(arg)
+        elif opt in ('-w', '--warnings'):
+            if arg is None or arg == '':
+                warnings = None
+            elif arg.lower() in ('yes', 'true', '1'):
+                warnings = True
+            elif arg.lower() in ('no', 'false', '0'):
+                warnings = False
+            else:
+                sys.exit('cli4: --warnings takes boolean True/False argument')
         elif opt in ('-d', '--dump'):
-            dump = True
-        elif opt in ('-a', '--api'):
-            dump_from_web = True
+            do_dump = True
         elif opt in ('-A', '--openapi'):
-            openapi_url = arg
+            do_openapi = True
+            openapi_url = arg if arg != '' else None
         elif opt in ('-b', '--binary'):
             binary_file = True
         elif opt in ('-G', '--get'):
@@ -437,23 +491,30 @@ def do_it(args):
         elif opt in ('-D', '--delete'):
             method = 'DELETE'
 
+    if example:
+        try:
+            examples.display()
+        except ModuleNotFoundError as e:
+            sys.exit(e)
+        sys.exit(0)
+
     try:
-        cf = CloudFlare.CloudFlare(debug=verbose, raw=raw, profile=profile)
+        cf = CloudFlare.CloudFlare(debug=verbose, raw=raw, profile=profile, http_headers=http_headers, warnings=warnings)
     except Exception as e:
         sys.exit(e)
 
-    if dump:
+    if do_dump:
         a = dump_commands(cf)
+        # success - just dump results and exit
         sys.stdout.write(a)
         sys.exit(0)
 
-    if dump_from_web:
-        a = dump_commands_from_web(cf)
-        sys.stdout.write(a)
-        sys.exit(0)
-
-    if openapi_url:
-        a = dump_commands_from_web(cf, openapi_url)
+    if do_openapi:
+        try:
+            a = dump_commands_from_web(cf, openapi_url)
+        except CloudFlare.exceptions.CloudFlareAPIError as e:
+            sys.exit('cli4: %s - Failed' % (e))
+        # success - just dump results and exit
         sys.stdout.write(a)
         sys.exit(0)
 
